(function (global) {
    "use strict";

    global.FRCC = global.FRCC || {};
    global.FRCC.AmeriCorps = global.FRCC.AmeriCorps || {};
    global.FRCC.AmeriCorps.Modals = (function () {
        const NegotiatedIndirectCostRateAgreement = (function () {
            const modalIdentifier = "#eyga-negotiated-indirect-cost-rate-agreement";
            const modalSelector = `${modalIdentifier}-modal`;
            const modalTitleSelector = `${modalIdentifier}-modal-title`;
            const modalSubmitSelector = `${modalIdentifier}-submit`;

            let isUpdate = null;
            let callingContext = null;

            const RateAgreementStatuses = {
                PendingNegotiation: 643260000,
                Active: 643260001,
                Rejected: 643260002,
                Expired: 643260003,
                Inactive: 643260004,
            };

            const RateAgreementStateCodes = {
                Active: 0,
                Inactive: 1,
            };

            const RateAgreementStatusCodes = {
                Active: 1,
                Inactive: 2,
            };

            const openModal = function (organizationId, negotiatedIndirectCostRateAgreementId, context = "Grid") {
                callingContext = context;

                resetModal(negotiatedIndirectCostRateAgreementId);
                $(modalSelector).modal("show");

                $(modalSubmitSelector)
                    .off("click")
                    .on("click", function () {
                        createNegotiatedIndirectCostRateAgreementRequest(organizationId, negotiatedIndirectCostRateAgreementId);
                    });
            };

            const resetModal = function (negotiatedIndirectCostRateAgreementId) {
                isUpdate = !!negotiatedIndirectCostRateAgreementId;

                const setTitle = isUpdate ? "Edit" : "Add";
                $(modalTitleSelector).text(`${setTitle} Indirect Cost Rate`);

                if (!isUpdate) {
                    $(`${modalSelector} .form-control`).val(null);
                    return;
                }

                const negotiatedIndirectCostRateAgreement = negotiatedIndirectCostRateAgreements.filter(
                    (x) => x.Id === negotiatedIndirectCostRateAgreementId
                )[0];
                if (!negotiatedIndirectCostRateAgreement) {
                    return;
                }

                $(`${modalIdentifier}-date`).val(Utilities.formatDateInput(negotiatedIndirectCostRateAgreement.RateAgreementDate));
                $(`${modalIdentifier}-cognizant-agency`).val(negotiatedIndirectCostRateAgreement.CognizantAgencyId);
                $(`${modalIdentifier}-fringe-benefits`).val(negotiatedIndirectCostRateAgreement.TreatmentOfFringeBenefits.replaceAll("\\n", "\n"));
                $(`${modalIdentifier}-paid-absences`).val(negotiatedIndirectCostRateAgreement.TreatmentOfPaidAbsences.replaceAll("\\n", "\n"));
                $(`${modalIdentifier}-base`).val(negotiatedIndirectCostRateAgreement.Base.replaceAll("\\n", "\n"));
                $(`${modalIdentifier}-additional-details`).val(negotiatedIndirectCostRateAgreement.AdditionalDetails.replaceAll("\\n", "\n"));
            };

            const createNegotiatedIndirectCostRateAgreementRequest = function (organizationId, negotiatedIndirectCostRateAgreementId) {
                if (!Validations.validateForm(modalSelector)) {
                    return;
                }

                const rateAgreementDate = $(`${modalIdentifier}-date`).val();
                const cognizantAgencyId = $(`${modalIdentifier}-cognizant-agency`).val();
                const fringeBenefits = $(`${modalIdentifier}-fringe-benefits`).val();
                const paidAbsences = $(`${modalIdentifier}-paid-absences`).val();
                const base = $(`${modalIdentifier}-base`).val();
                const additionalDetails = $(`${modalIdentifier}-additional-details`).val();
                const rateAgreementStatus = RateAgreementStatuses.Active;

                const payload = JSON.stringify({
                    ["eyfrcc_organization@odata.bind"]: `/accounts(${organizationId})`,
                    ["eyfrcc_CognizantAgency@odata.bind"]: `/eyfrcc_governmentalunits(${cognizantAgencyId})`,
                    ["eyfrcc_rateagreementdate"]: rateAgreementDate,
                    ["eyfrcc_rateagreementstatus"]: rateAgreementStatus,
                    ["eyfrcc_treatmentoffringebenefits"]: fringeBenefits,
                    ["eyfrcc_treatmentofpaidabsences"]: paidAbsences,
                    ["eyfrcc_base"]: base,
                    ["eyfrcc_additionaldetails"]: additionalDetails,
                });

                upsertNegotiatedIndirectCostRateAgreement(payload, negotiatedIndirectCostRateAgreementId);
            };

            const upsertNegotiatedIndirectCostRateAgreement = function (payload, negotiatedIndirectCostRateAgreementId) {
                const requestType = !negotiatedIndirectCostRateAgreementId ? "POST" : "PATCH";
                const requestUrl = !negotiatedIndirectCostRateAgreementId
                    ? "/_api/eyfrcc_indirectcostrates"
                    : `/_api/eyfrcc_indirectcostrates(${negotiatedIndirectCostRateAgreementId})`;

                Utilities.sendAjaxRequest(
                    {
                        type: requestType,
                        url: requestUrl,
                        data: payload,
                        success: async function (res, status, xhr) {
                            handleNegotiatedIndirectCostRateAgreementsPage(1);
                            if (negotiatedIndirectCostRateAgreementId && callingContext !== "Grid") {
                                displayNegotiatedIndirectCostRateAgreement(negotiatedIndirectCostRateAgreementId, RateAgreementStateCodes.Active);
                            }
                            FRCC.Portal.Utilities.hideProgressIndicator();
                            $(modalSelector).modal("hide");
                        },
                    },
                    `${isUpdate ? "Updating" : "Creating"} Negotiated Cost Rate Agreement...`
                );
            };

            const deactivateConfirmation = function (negotiatedIndirectCostRateAgreementId) {
                Confirmation.openModal({
                    Message: "Please confirm that you want to deactivate this Negotiated Indirect Cost Rate Agreement",
                    ConfirmationCallback: () => deactivateNegotiatedIndirectCostRateAgreement(negotiatedIndirectCostRateAgreementId),
                });
            };

            const deactivateNegotiatedIndirectCostRateAgreement = function (negotiatedIndirectCostRateAgreementId) {
                const requestUrl = `/_api/eyfrcc_indirectcostrates(${negotiatedIndirectCostRateAgreementId})`;

                Utilities.sendAjaxRequest(
                    {
                        type: "PATCH",
                        url: requestUrl,
                        contentType: "application/json",
                        data: JSON.stringify({
                            ["statecode"]: RateAgreementStateCodes.Inactive,
                            ["statuscode"]: RateAgreementStatusCodes.Inactive,
                            ["eyfrcc_rateagreementstatus"]: RateAgreementStatuses.Inactive,
                        }),
                        success: () => {
                            handleNegotiatedIndirectCostRateAgreementsPage(1);
                            FRCC.Portal.Utilities.hideProgressIndicator();
                            Confirmation.closeModal();
                        },
                    },
                    "Deactivating Negotiated Indirect Cost Rate Agreement..."
                );
            };

            return {
                openModal: openModal,
                resetModal: resetModal,
                deactivateConfirmation: deactivateConfirmation,
            };
        })();

        const RateLine = (function () {
            const modalIdentifier = "#eyga-rate-line";
            const modalSelector = `${modalIdentifier}-modal`;
            const modalTitleSelector = `${modalIdentifier}-modal-title`;
            const modalSubmitSelector = `${modalIdentifier}-submit`;

            let isUpdate = null;
            let entityLogicalCollectionName = null;
            let entityDisplayName = null;
            let isFringeBenefit = null;
            let rateLines = null;
            let loadingLabel = null;
            let callbackHandler = null;

            const openModal = function (negotiatedIndirectCostRateAgreementId, rateLineId, logicalName) {
                setEntityVariables(rateLineId, logicalName);

                resetModal(rateLineId);
                $(modalSelector).modal("show");

                $(modalSubmitSelector)
                    .off("click")
                    .on("click", function () {
                        createRateLineRequest(negotiatedIndirectCostRateAgreementId, rateLineId);
                    });

                $(`${modalIdentifier}-until-amended`).off("change").on("change", clearAndDisableToDate);
            };

            const setEntityVariables = function (rateLineId, logicalName) {
                isUpdate = !!rateLineId;
                isFringeBenefit = logicalName === "eyfrcc_fringebenefitratelines";

                entityLogicalCollectionName = FRCC.Portal.Utilities.getEntitySetName(logicalName);
                entityDisplayName = isFringeBenefit ? "Fringe Benefit Rate" : "Indirect Cost Rate";
                rateLines = isFringeBenefit ? fringeBenefitRateLines : indirectCostRateLines;
                callbackHandler = isFringeBenefit ? handleFringeBenefitRateLinesPage : handleIndirectCostRateLinesPage;

                loadingLabel = isUpdate ? `Updating ${entityDisplayName}...` : `Creating ${entityDisplayName}...`;
            };

            const resetModal = function (rateLineId) {
                $(`${modalIdentifier} .modal-body`).scrollTop(0);

                const setTitle = isUpdate ? "Edit" : "Add";
                $(modalTitleSelector).text(`${setTitle} ${entityDisplayName}`);

                if (!isUpdate) {
                    $(`${modalSelector} .form-control`).val(null);
                    $(`${modalIdentifier}-until-amended`).prop("checked", false);
                    clearAndDisableToDate();
                    return;
                }

                const rateLine = rateLines.filter((x) => x.Id === rateLineId)[0];
                if (!rateLine) {
                    return;
                }

                $(`${modalIdentifier}-type`).val(rateLine.RateTypeValue);
                $(`${modalIdentifier}-from-date`).val(Utilities.formatDateInput(rateLine.FromDate));
                $(`${modalIdentifier}-until-amended`).prop("checked", rateLine.UntilAmended);
                $(`${modalIdentifier}-to-date`).val(Utilities.formatDateInput(rateLine.ToDate));
                $(`${modalIdentifier}-rate`).val(rateLine.RatePercent);
                $(`${modalIdentifier}-location`).val(rateLine.Location);
                $(`${modalIdentifier}-applicable-to`).val(rateLine.ApplicableTo);

                clearAndDisableToDate();
            };

            const createRateLineRequest = function (negotiatedIndirectCostRateAgreementId, rateLineId) {
                if (!Validations.validateForm(modalSelector)) {
                    return;
                }

                const rateAgreementType = $(`${modalIdentifier}-type`).val();
                const fromDate = $(`${modalIdentifier}-from-date`).val();
                const untilAmended = $(`${modalIdentifier}-until-amended`).prop("checked");
                const toDate = $(`${modalIdentifier}-to-date`).val();
                const rate = $(`${modalIdentifier}-rate`).val();
                const location = $(`${modalIdentifier}-location`).val();
                const applicableTo = $(`${modalIdentifier}-applicable-to`).val();

                const payload = JSON.stringify({
                    ["eyfrcc_ratetype"]: rateAgreementType,
                    ["eyfrcc_fromdate"]: fromDate,
                    ["eyfrcc_todate"]: !toDate ? null : toDate,
                    [isFringeBenefit ? "eyfrcc_untilamendednew" : "eyfrcc_untilamended"]: untilAmended,
                    ["eyfrcc_rate"]: !parseFloat(rate) ? null : parseFloat(rate),
                    ["eyfrcc_location"]: location,
                    ["eyfrcc_applicableto"]: applicableTo,
                    ["eyfrcc_IndirectCostRateId@odata.bind"]: `/eyfrcc_indirectcostrates(${negotiatedIndirectCostRateAgreementId})`,
                });

                upsertRateLine(payload, rateLineId);
            };

            const upsertRateLine = function (payload, rateLineId) {
                const requestType = !isUpdate ? "POST" : "PATCH";
                const requestUrl = !isUpdate ? `/_api/${entityLogicalCollectionName}` : `/_api/${entityLogicalCollectionName}(${rateLineId})`;

                Utilities.sendAjaxRequest(
                    {
                        type: requestType,
                        url: requestUrl,
                        data: payload,
                        success: async function (res, status, xhr) {
                            callbackHandler(1);
                            FRCC.Portal.Utilities.hideProgressIndicator();
                            $(modalSelector).modal("hide");
                        },
                    },
                    loadingLabel
                );
            };

            const clearAndDisableToDate = function () {
                const untilAmended = $(`${modalIdentifier}-until-amended`).prop("checked");
                $(`${modalIdentifier}-to-date`).prop("disabled", untilAmended);

                if (untilAmended) {
                    $(`${modalIdentifier}-to-date`).val(null);
                }
            };

            const deleteConfirmation = function (rateLineId, logicalName) {
                setEntityVariables(rateLineId, logicalName);

                Confirmation.openModal({
                    Message: `Please confirm that you want to delete this ${entityDisplayName}`,
                    ConfirmationCallback: () => deleteRateLine(rateLineId),
                });
            };

            const deleteRateLine = function (rateLineId) {
                const requestUrl = `/_api/${entityLogicalCollectionName}(${rateLineId})`;

                Utilities.sendAjaxRequest(
                    {
                        type: "DELETE",
                        url: requestUrl,
                        contentType: "application/json",
                        success: () => {
                            callbackHandler(1);
                            FRCC.Portal.Utilities.hideProgressIndicator();
                            Confirmation.closeModal();
                        },
                    },
                    loadingLabel
                );
            };

            return {
                openModal: openModal,
                resetModal: resetModal,
                deleteConfirmation: deleteConfirmation,
            };
        })();

        const Project = (function () {
            const ProjectStatuses = {
                ProjectRequestReview: 643260000,
                ProjectAvailable: 643260001,
                ProjectAwarded: 643260002,
                ProjectDuplicate: 643260003,
                ProjectClosed: 643260004,
                ProjectDraft: 643260005,
            };

            const checkProjectState = async function (organizationId, programId, stateId) {
                if (!organizationId || !programId || !stateId) {
                    return;
                }

                const projectsResponse = await FRCC.Portal.Utilities.asyncAppAjax({
                    type: "GET",
                    url: `/_api/eyfrcc_projects?$filter=_eyfrcc_program_value eq '${programId}' and _eyfrcc_projectaddressstateterritory_value eq '${stateId}' and _eyfrcc_organizationid_value eq '${organizationId}'`,
                    contentType: "application/json",
                });

                let flag = false;
                projectsResponse.value.forEach((val) => {
                    if (val.eyfrcc_projectstatus !== ProjectStatuses.ProjectDuplicate && val.eyfrcc_projectstatus !== ProjectStatuses.ProjectClosed) {
                        flag = true;
                    }
                });

                $("#eyga-project-state-warning").toggle(flag);
            };

            return {
                checkProjectState: checkProjectState,
            };
        })();

        const SessionManagement = (function () {
            const modalIdentifier = "#eyga-session-management";
            const modalSelector = `${modalIdentifier}-modal`;
            const modalBodyErrorContainerSelector = `${modalSelector} .modal-body-error-container`;
            const modalSessionWarningSelector = ".modal-body-session-warning";
            const modalCountdownTimerBodySelector = ".modal-body-countdown-timer";
            const modalCountdownTimerSelector = ".modal-body-countdown-timer span";
            const modalLogoutSelector = `${modalIdentifier}-logout`;
            const modalExtendSessionSelector = `${modalIdentifier}-extend-session`;

            const openWarningModal = function (sessionWarningThreshold, sessionTimeoutThreshold) {
                resetWarningModal(sessionWarningThreshold, sessionTimeoutThreshold);
                $(modalSelector).modal("show");

                $(modalLogoutSelector)
                    .off("click")
                    .on("click", function () {
                        logout();
                    });

                $(modalExtendSessionSelector)
                    .off("click")
                    .on("click", function () {
                        extendSession();
                    });
            };

            const resetWarningModal = function (sessionWarningThreshold, sessionTimeoutThreshold) {
                const minutesUntilTimeout = Math.ceil(sessionTimeoutThreshold / 60 / 1000 - sessionWarningThreshold / 60 / 1000);
                $(modalSessionWarningSelector).text(
                    `Your session will expire in ${minutesUntilTimeout} minute(s) due to inactivity. All unsaved changes will be lost.`
                );
                $(modalCountdownTimerBodySelector).hide();
                $(modalBodyErrorContainerSelector).show();
            };

            const displayCountdown = function () {
                $(modalSessionWarningSelector).text(`Your session will expire in 1 minute due to inactivity. All unsaved changes will be lost.`);
                $(modalCountdownTimerBodySelector).show();

                setInterval(updateCountdown, 1000);
            };

            // TO-DO: Improve countdown logic
            // Right now will only update the modal when it reaches one minute
            // If warning was 5 minutes and timeout 10 minutes for example, would be nice if modal message updated to 4, 3, 2, then 1
            let countdownDuration = 60;
            const updateCountdown = function () {
                const minutes = Math.floor(countdownDuration / 60);
                const seconds = countdownDuration % 60;

                const formattedTime = `${minutes}:${seconds.toString().padStart(2, "0")}`;
                $(modalCountdownTimerSelector).text(formattedTime);

                if (countdownDuration > 0) {
                    countdownDuration--;
                }
            };

            const closeWarningModal = function () {
                $(modalSelector).modal("hide");
            };

            const extendSession = function () {
                FRCC.AmeriCorps.SessionManagement.extendSession();
                closeWarningModal();
            };

            const openTimeoutModal = function (returnUrl) {
                Confirmation.openModal({
                    Title: "Session Expired",
                    Message: "Your session has expired due to inactivity. Please sign in again to continue.",
                    CancelButtonLabel: "Close",
                    ConfirmButtonLabel: "Sign In",
                    NotificationLevel: "Warning",
                    ConfirmationCallback: () => signInUser(returnUrl),
                });
            };

            const signInUser = function (returnUrl) {
                window.location.href = `${window.location.origin}/Account/Login/?returnUrl=${returnUrl || "/"}`;
            };

            return {
                openWarningModal: openWarningModal,
                closeWarningModal: closeWarningModal,
                openTimeoutModal: openTimeoutModal,
                displayCountdown: displayCountdown,
            };
        })();

        const LateSubmission = (function () {
            const modalIdentifier = "#late-submission";
            const modalSelector = `${modalIdentifier}-modal`;

            const openModal = function () {
                return new Promise((resolve, reject) => {
                    resetModal();
                    $(modalSelector).modal("show");

                    $(`${modalSelector} .cancel`)
                        .off("click")
                        .on("click", function () {
                            resolve({
                                Success: false,
                                EncounteredFault: false,
                            });
                        });

                    $(`${modalIdentifier}-submit`)
                        .off("click")
                        .on("click", function () {
                            updateRecord();
                            closeModal();
                            resolve({
                                Success: true,
                            });
                        });
                });
            };

            const closeModal = function () {
                $(modalSelector).modal("hide");
            };

            const resetModal = function () {
                //set modal options to orig state
            };

            const updateRecord = function () {
                var reasonForLateSubmission = $("#late-submission-reasonforlatesubmission").val();
                var data = {
                  eyfrcc_latesubmissionjustification: reasonForLateSubmission
                };

                updateApplication(data, function () {
                    $('#late-submission-modal').modal('hide');
                    $('input.submit-btn.btn.btn-primary').last().click();
                });    
                setTimeout(()=> {},1000);          
            };

            const getEntitySetName = function (logicalName) {            
                if (!logicalName) { return logicalName; }
                if (logicalName.endsWith('s')) { return logicalName + 'es'; }
                if (logicalName.endsWith('y')) { return logicalName + 'ies'; }
                return logicalName + 's';
            };

            const updateApplication = function (data, callback) {    
                const regardingId = $("#hdn-recordid").val();
                const regardingType = $("#hdn-recordtype").val();
    
                var entityLogicalName = getEntitySetName(regardingType);
                var ajaxOptions = {
                    type: "PATCH",
                    url: ["/_api/", entityLogicalName, "(", regardingId, ")"].join(''),
                    contentType: "application/json",
                    data: JSON.stringify(data),
                    success: function(res, status, xhr) {
                        console.log("eyfrcc_latesubmissionjustification updated");
                        callback();
                    }
                };
                var message = "Submitting Application...";
                FRCC.Portal.Utilities.appAjax(message, ajaxOptions)
                    .fail(function(response) {
                        if (response.responseJSON) {
                            alert("Error: " + response.responseJSON.error.message);
                        } else {
                            alert("Error: Web API is not available... ");
                        }
                });
            };

            return {
                openModal: openModal,
                closeModal: closeModal,
            };
        })();

        const Confirmation = (function () {
            const modalIdentifier = "#eyga-confirm";
            const modalSelector = `${modalIdentifier}-modal`;
            const modalTitleSelector = `${modalIdentifier}-modal-title`;
            const modalSubmitSelector = `${modalIdentifier}-submit`;
            const modalDefaultErrorAlertContainerSelector = `${modalSelector} .modal-body-error-container`;

            const DefaultConfiguration = {
                Message: "",
                MessageContainsHTML: false,
                Content: null,
                ShowDefaultError: false,
                Title: "Confirmation",
                CancelButtonLabel: "Cancel",
                ConfirmButtonLabel: "Confirm",
                ShowCancelButton: true,
                ShowConfirmationButton: true,
                NotificationLevel: "None",
                ModalSize: "Small",
                ConfirmationCallback: null,
            };

            const NotificationClasses = {
                Warning: "alert alert-warning",
                Error: "alert alert-danger",
                Info: "alert alert-info",
                Success: "alert alert-success",
                None: "",
            };

            const ModalSizeClasses = {
                Small: "modal-sm",
                Medium: "modal-md",
                Large: "modal-lg",
                ExtraLarge: "modal-xl",
            };

            const openModal = function (configuration = {}) {
                const fullConfiguration = mergeConfigurations(DefaultConfiguration, configuration);

                resetModal(fullConfiguration);

                $(modalSelector).modal("show");
            };

            const openModalWithPromises = function (configuration) {
                openModal(configuration);

                return new Promise((resolve, reject) => {
                    $(modalSelector).modal("show");

                    $(`${modalSelector} .cancel`)
                        .off("click")
                        .on("click", function () {
                            closeModal();
                            resolve({
                                Success: false,
                                EncounteredFault: false,
                            });
                        });

                    $(`${modalIdentifier}-submit`)
                        .off("click")
                        .on("click", function () {
                            closeModal();
                            resolve({
                                Success: true,
                            });
                        });
                });
            };

            const mergeConfigurations = function (defaultConfig, userConfig) {
                const cleanedUserConfig = cleanObject(userConfig);
                return { ...defaultConfig, ...cleanedUserConfig };
            };

            const cleanObject = function (obj) {
                return Object.fromEntries(Object.entries(obj).filter(([key, value]) => value !== undefined));
            };

            const resetModal = function (configuration) {
                const $modal = $(modalSelector);
                const $modalDialog = $modal.find(".modal-dialog");
                const $modalBodyDescription = $modal.find(".modal-body-description");
                const $modalBodyContent = $modal.find(".modal-body-content");
                const $cancelButton = $modal.find(".cancel");
                const $submitButton = $(modalSubmitSelector);

                $modal.find(".modal-title").text(configuration.Title);

                if (configuration.ShowDefaultError) {
                    $(modalDefaultErrorAlertContainerSelector).show();
                    $modalBodyDescription.hide();
                    $modalBodyContent.hide();
                } else {
                    if (configuration.MessageContainsHTML) {
                        $modalBodyDescription.html(configuration.Message);
                    } else {
                        $modalBodyDescription.text(configuration.Message);
                    }

                    $modalBodyDescription
                        .removeClass("alert alert-warning alert-info alert-danger")
                        .addClass(retrieveNotificationClasses(configuration.NotificationLevel))
                        .show();

                    if (configuration.Content) {
                        $modalBodyContent.empty().append(configuration.Content);
                        $modalBodyContent.show();
                    }

                    $(modalDefaultErrorAlertContainerSelector).hide();
                }

                $modalDialog.removeClass("modal-sm modal-md modal-lg modal-xl").addClass(retrieveNotificationClasses(configuration.ModalSize));
                $cancelButton.text(configuration.CancelButtonLabel).toggle(configuration.ShowCancelButton);
                $submitButton.text(configuration.ConfirmButtonLabel).toggle(configuration.ShowConfirmationButton).off("click");

                if (configuration.ConfirmationCallback) {
                    $submitButton.on("click", configuration.ConfirmationCallback);
                }
            };

            const retrieveNotificationClasses = function (notificationLevel) {
                return NotificationClasses[notificationLevel] || "";
            };

            const retrieveModalSizeClasses = function (modalSize) {
                return ModalSizeClasses[modalSize] || "modal-sm";
            };

            const closeModal = function (forwardToUrl, loadingMessage) {
                $(modalSelector).modal("hide");
                if(forwardToUrl){
                    if(loadingMessage){
                        FRCC.Portal.Utilities.showProgressIndicator(loadingMessage);
                    }
                    window.location.href = forwardToUrl;
                }
            };

            return {
                openModal: openModal,
                openModalWithPromises: openModalWithPromises,
                resetModal: resetModal,
                closeModal: closeModal,
            };
        })();

        const AddressValidation = (function () {
            const modalSelector = "#addressModal";
            const modalTitleSelector = "#addressModalLabel";

            const modalCancelSelector = "#addressModalCancelBtn";
            const modalUseOriginalSelector = "#addressModalNoChangesBtn";
            const modalUseSuggestedSelector = "#addressModalSaveBtn";

            const openModal = function (request, response) {
                return new Promise((resolve, reject) => {
                    const originalAddressFormatted = formatAddressLines(request);
                    $("#original-address-formatted").html(originalAddressFormatted);

                    const suggestedAddressFormatted = formatAddressLines(response.Address, request);
                    $("#suggested-address-formatted").html(suggestedAddressFormatted);

                    $(modalCancelSelector)
                        .off("click")
                        .on("click", () => {
                            reject();
                        });

                    $(modalUseOriginalSelector)
                        .off("click")
                        .on("click", () => {
                            resolve();
                        });

                    $(modalUseSuggestedSelector)
                        .off("click")
                        .on("click", () => {
                            resolve(response.Address);
                        });

                    $(modalSelector).modal("show");
                });
            };

            const resetModal = function (confirmationMessage) {
                $(`${modalSelector} .modal-body-description`).text(confirmationMessage);
            };

            const closeModal = function () {
                $(modalSelector).modal("hide");
            };

            const validateAddress = async function (userId, selectors) {
                const addressValidationRequest = {
                    AddressLines: [$(selectors.StreetAddress1).val(), $(selectors.StreetAddress2).val()],
                    Locality: $(selectors.City).val(),
                    AdministrativeArea: extractStateAbbreviation($(`${selectors.State} option:selected`).text()),
                    PostalCode: $(selectors.PostalCode).val(),
                    RegionCode: "US",
                };

                FRCC.Portal.Utilities.showProgressIndicator("Validating Address...");

                const eygaApiContext = new FRCC.Portal.Utilities.EygaApiContext(userId, "contact", userId);
                const response = await FRCC.Portal.Utilities.callEygaApi(
                    FRCC.Portal.Utilities.ApiCall.ValidateAddress,
                    eygaApiContext,
                    addressValidationRequest
                );

                FRCC.Portal.Utilities.hideProgressIndicator();
                if (response.Address && response.Address.AddressLines) {
                    response.Address.AddressLines = response.Address.AddressLines.map((line) => (line === null ? "" : line));
                }

                if (response.Address && response.Address.AddressLines) {
                    if (response.Address.PostalCode.length === 6 && response.Address.PostalCode.charAt(5) === "-") {
                        response.Address.PostalCode = response.Address.PostalCode + "0000";
                    }
                }

                if (response.ValidationLevel == "Invalid" || !response.Address) {
                    const $address = $(formatAddressLines(addressValidationRequest));
                    $address.css("margin-bottom", "0px").css("font-weight", "bold");

                    const configuration = {
                        Message: `The address entered could not be verified by the US Postal Service. Please review and confirm the address is accurate.
                            <br>
                            <br>
                            ${$address.prop("outerHTML")}`,
                        MessageContainsHTML: true,
                        Title: "Address Not Found",
                        ConfirmButtonLabel: "Continue",
                        ShowCancelButton: true,
                        ShowConfirmationButton: true,
                        CancelButtonLabel: "Edit Address",
                        ConfirmButtonLabel: "Use Address As Entered",
                        NotificationLevel: "Warning",
                        ModalSize: "Medium",
                    };

                    const confirmationResult = await Confirmation.openModalWithPromises(configuration);
                    if (confirmationResult?.Success) {
                        return true;
                    }

                    $(selectors.StreetAddress1).focus();
                    return null;
                } else if (
                    addressValidationRequest.AddressLines.every(
                        (x, i) => x?.trim()?.toLowerCase() == response.Address.AddressLines?.[i]?.trim()?.toLowerCase()
                    ) &&
                    addressValidationRequest.Locality?.trim()?.toLowerCase() == response.Address.Locality?.trim()?.toLowerCase() &&
                    addressValidationRequest.AdministrativeArea?.trim()?.toLowerCase() == response.Address.AdministrativeArea?.trim()?.toLowerCase() &&
                    addressValidationRequest.PostalCode?.trim()?.toLowerCase() == response.Address.PostalCode?.trim()?.toLowerCase() &&
                    addressValidationRequest.RegionCode?.trim()?.toLowerCase() == response.Address.RegionCode?.trim()?.toLowerCase()
                ) {
                    return true;
                }

                return new Promise((resolve, reject) => {
                    openModal(addressValidationRequest, response).then(
                        function accept(address) {
                            if (address) {
                                $(selectors.StreetAddress1).val(address.AddressLines[0]);
                                $(selectors.StreetAddress2).val(address.AddressLines[1]);
                                $(selectors.City).val(address.Locality);
                                selectStateByAbbreviation(selectors.State, address.AdministrativeArea);
                                $(selectors.PostalCode).val(address.PostalCode);
                            }

                            closeModal();
                            resolve(true);
                        },
                        function cancel() {
                            resolve(null);
                        }
                    );
                });
            };

            //Address validation when the state field is a lookup
            const validateAddressLookup = async function (userId, selectors, successCallback, window$ = null) {
                const stateAbr = function (stateName) {
                    if (!stateName) {
                        return;
                    }
                    return new Promise(function (resolve, reject) {
                        FRCC.Portal.Utilities.appAjax("Get State", {
                            type: "Get",
                            url: `/_api/eyfrcc_stateterritories?$filter=eyfrcc_stateterritoryname%20eq%20%27${stateName}%27&$select=eyfrcc_abbreviation`,
                            contentType: "application/json",
                            success: function (response) {
                                if (response) {
                                    console.log("response: " + response.value[0].eyfrcc_abbreviation);
                                    resolve(response.value[0].eyfrcc_abbreviation);
                                } else {
                                    reject("not found");
                                }
                            },
                            error: function (xhr, status, error) {
                                reject(error);
                            },
                        });
                    });
                };

                async function main() {
                    try {
                        const state = await stateAbr(window$(selectors.State).val());
                        const addressValidationRequest = {
                            AddressLines: [window$(selectors.StreetAddress1).val(), window$(selectors.StreetAddress2).val()],
                            Locality: window$(selectors.City).val(),
                            AdministrativeArea: state,
                            PostalCode: window$(selectors.PostalCode).val(),
                            RegionCode: "US",
                        };
                        FRCC.Portal.Utilities.showProgressIndicator("Validating Address...");

                        const eygaApiContext = new FRCC.Portal.Utilities.EygaApiContext(userId, "contact", userId);
                        const response = await FRCC.Portal.Utilities.callEygaApi(
                            FRCC.Portal.Utilities.ApiCall.ValidateAddress,
                            eygaApiContext,
                            addressValidationRequest
                        );

                        FRCC.Portal.Utilities.hideProgressIndicator();

                        if (response.ValidationLevel == "Invalid") {
                            if ($(".modal-body-error-container").length > 0) {
                                return;
                            } else {
                                $('<div class="modal-body-error-container alert alert-danger" role="alert" aria-live="assertive"></div>').insertAfter(
                                    ".form-loading"
                                );
                                $(".modal-body-error-container").append('<p class="modal-body-description">The address you have entered is invalid.</p>');
                                $(".modal-body-error-container").show();
                            }

                            return false;
                        } else if (
                            addressValidationRequest.AddressLines.every(
                                (x, i) => x?.trim()?.toLowerCase() == response.Address.AddressLines?.[i]?.trim()?.toLowerCase()
                            ) &&
                            addressValidationRequest.Locality?.trim()?.toLowerCase() == response.Address.Locality?.trim()?.toLowerCase() &&
                            addressValidationRequest.AdministrativeArea?.trim()?.toLowerCase() == response.Address.AdministrativeArea?.trim()?.toLowerCase() &&
                            addressValidationRequest.PostalCode?.trim()?.toLowerCase() == response.Address.PostalCode?.trim()?.toLowerCase() &&
                            addressValidationRequest.RegionCode?.trim()?.toLowerCase() == response.Address.RegionCode?.trim()?.toLowerCase()
                        ) {
                            successCallback();
                            return;
                        }

                        return openModal(addressValidationRequest, response).then(
                            function accept(address) {
                                if (address) {
                                    $(selectors.StreetAddress1).val(address.AddressLines[0]);
                                    $(selectors.StreetAddress2).val(address.AddressLines[1]);
                                    $(selectors.City).val(address.Locality);
                                    $(selectors.State).val(address.AdministrativeArea);
                                    $(selectors.PostalCode).val(address.PostalCode);
                                }

                                //Utilities.clearValidationErrors();
                                $(".modal-body-error-container").remove();
                                closeModal();
                                successCallback();
                            },
                            function cancel() {
                                return;
                            }
                        );
                    } catch (error) {
                        console.error("error: " + error);
                    }
                }
                main();
            };

            const formatAddressLines = function (address, comparisonAddress = null) {
                const street1 = formatAddressLine(address.AddressLines?.[0], comparisonAddress?.AddressLines?.[0]);
                const street2 = formatAddressLine(address.AddressLines?.[1], comparisonAddress?.AddressLines?.[1]);
                const city = formatAddressLine(address.Locality, comparisonAddress?.Locality);
                const stateCode = formatAddressLine(address.AdministrativeArea, comparisonAddress?.AdministrativeArea);
                const postalCode = formatAddressLine(address.PostalCode, comparisonAddress?.PostalCode);

                return `<address>${street1}</br>${street2}${street2 ? "</br>" : ""}${city}, ${stateCode} ${postalCode}</address>`;
            };

            const formatAddressLine = function (addressLine, comparisonLine) {
                if (!addressLine) {
                    return "";
                }

                if (!comparisonLine && comparisonLine != "") {
                    return addressLine;
                }

                const isDifferent = comparisonLine?.toLowerCase() !== addressLine?.toLowerCase();

                return isDifferent ? `<em class="text-danger text-weight-semibold">${addressLine}</em>` : addressLine;
            };

            const extractStateAbbreviation = function (fullStateText) {
                if (!fullStateText || !fullStateText.includes(",")) {
                    return null;
                }

                const parts = fullStateText.split(",");
                if (parts.length !== 2) {
                    return null;
                }

                return parts[1].trim();
            };

            const selectStateByAbbreviation = function (idSelector, stateAbbreviation) {
                $(`${idSelector} option`).each(function () {
                    const parts = $(this)
                        .text()
                        .split(",")
                        .map((part) => $.trim(part));
                    if (parts.length === 2 && parts[1] === stateAbbreviation) {
                        $(this).parent().val($(this).val());
                        return false;
                    }
                });
            };

            return {
                openModal: openModal,
                resetModal: resetModal,
                closeModal: closeModal,
                validateAddress: validateAddress,
                validateAddressLookup: validateAddressLookup,
            };
        })();

        const Messages = (function () {
            const ModalIdentifiers = {
                CreateMessage: "#eyga-create-message",
                ViewMessage: "#eyga-view-message",
            };

            const ModalSelectors = {
                CreateMessage: {
                    Modal: `${ModalIdentifiers.CreateMessage}-modal`,
                    Title: `${ModalIdentifiers.CreateMessage}-title`,
                    Form: `${ModalIdentifiers.CreateMessage}-form`,
                    Subject: `${ModalIdentifiers.CreateMessage}-subject`,
                    Content: `${ModalIdentifiers.CreateMessage}-content`,
                    FileInput: `${ModalIdentifiers.CreateMessage}-file-input`,
                    Submit: `${ModalIdentifiers.CreateMessage}-submit`,
                    FileSuccessAlert: `${ModalIdentifiers.CreateMessage}-file-success-alert`,
                    FileErrorAlert: `${ModalIdentifiers.CreateMessage}-file-error-alert`,
                    FileUploading: `${ModalIdentifiers.CreateMessage}-file-uploading`,
                    FileTable: `${ModalIdentifiers.CreateMessage}-file-table`,
                    FileCategoryPrefix: `${ModalIdentifiers.CreateMessage}-file-category-`,
                    RemoveButton: ".remove-button",
                },
                ViewMessage: {
                    Modal: `${ModalIdentifiers.ViewMessage}-modal`,
                    DateSent: `${ModalIdentifiers.ViewMessage}-date-sent`,
                    Sender: `${ModalIdentifiers.ViewMessage}-sender`,
                    Regarding: `${ModalIdentifiers.ViewMessage}-regarding`,
                    Subject: `${ModalIdentifiers.ViewMessage}-modal .modal-body-subject`,
                    Description: `${ModalIdentifiers.ViewMessage}-modal .modal-body-description`,
                    DocumentContainer: `${ModalIdentifiers.ViewMessage}-modal .eyga-view-message-modal-document-container`,
                    DocumentButton: `${ModalIdentifiers.ViewMessage}-modal .eyga-view-message-modal-document-container button`,
                    ReplyButton: `${ModalIdentifiers.ViewMessage}-modal .reply-button`,
                },
            };

            const Senders = {
                Applicant: 643260000,
                AmeriCorps: 643260001,
            };

            let AttachedFiles = [];

            const openCreateMessageModal = async function (regardingId, regardingType, userId, subject, isReplying = false) {
                FRCC.Portal.Utilities.showContextualProgressIndicator("createMessage");

                resetCreateMessageModal(subject, isReplying);

                const applicationNumberResponse = await FRCC.Portal.Utilities.asyncAppAjax({
                    type: "GET",
                    url: `/Services/GetApplicationNumberBasedonApplicationType/?currentRecordId=${regardingId}`,
                    contentType: "application/json",
                });
                const applicationNumber = JSON.parse(applicationNumberResponse);
                const documentPath = applicationNumber.DocPath;
                const documentTag = applicationNumber.AppNumber;

                const documentMaxSize = await FRCC.Portal.Utilities.getEYGAConfigurationSetting("eyfrcc_maxfilesizemb");
                const documentValidExtensions = await FRCC.Portal.Utilities.getEYGAConfigurationSetting("eyfrcc_allowedextensions");

                FRCC.Portal.Utilities.hideContextualProgressIndicator("createMessage");

                $(ModalSelectors.CreateMessage.Submit)
                    .off("click")
                    .click(() => createMessageAndUploadFiles(documentPath, documentTag, regardingId, regardingType, userId));

                $(ModalSelectors.CreateMessage.FileInput)
                    .off("change")
                    .change(function () {
                        $(ModalSelectors.CreateMessage.FileSuccessAlert).hide();
                        $(ModalSelectors.CreateMessage.FileErrorAlert).hide();
                        $(ModalSelectors.CreateMessage.FileUploading).show();

                        const files = $(ModalSelectors.CreateMessage.FileInput).prop("files");
                        if (files) {
                            let validationErrors = [];
                            let validationWarnings = [];

                            for (let i = 0; i < files.length; i++) {
                                const file = files[i];
                                const validationError = FRCC.Portal.Documents.isFileValid(file, documentMaxSize, documentValidExtensions);
                                if (validationError) {
                                    validationErrors.push(validationError);
                                    continue;
                                }

                                const isFileDuplicate = AttachedFiles.some((f) => f.File.name === file.name);
                                if (isFileDuplicate) {
                                    validationWarnings.push(`The file ${file.name} has already been attached to the message.`);
                                    continue;
                                }

                                const fileId = Utilities.generateGuid();
                                AttachedFiles.push({
                                    Id: fileId,
                                    File: file,
                                });
                                const filepath = URL.createObjectURL(file);
                                $(`${ModalSelectors.CreateMessage.FileTable} tbody`).append(createFileRow(fileId, filepath, file.name));
                            }

                            $(ModalSelectors.CreateMessage.FileTable).toggle(AttachedFiles.length !== 0);
                            $(ModalSelectors.CreateMessage.FileUploading).hide();
                            handleFileAttachmentAlerts(validationErrors, validationWarnings);

                            $(ModalSelectors.CreateMessage.RemoveButton)
                                .off("click")
                                .click(function () {
                                    const fileId = $(this).closest("tr").attr("id");
                                    const index = AttachedFiles.map((f) => f.Id).indexOf(fileId);
                                    AttachedFiles.splice(index, 1);

                                    $(this).closest("tr").remove();
                                    if (AttachedFiles.length === 0) {
                                        $(ModalSelectors.CreateMessage.FileTable).hide();
                                    }
                                });
                        }
                    });

                $(ModalSelectors.CreateMessage.Form).show();
                $(isReplying ? ModalSelectors.CreateMessage.Content : ModalSelectors.CreateMessage.Subject).focus();
            };

            const createFileRow = function (i, filepath, fileName) {
                return `
                    <tr id="${i}">
                        <td>
                            <span class="text">${fileName}</span>
                        </td>
                        <td class="contains-component text-center">
                            <button type="button" aria-label="Remove file ${fileName}" class="btn btn-link btn-md btn-icon-only remove-button">
                                <span class="glyphicon glyphicon-trash icon-size-md"></span>
                            </button>
                        </td>
                    </tr>`;
            };

            const handleFileAttachmentAlerts = function (validationErrors, validationWarnings) {
                handleFileAttachmentErrorAlert(validationErrors);
                handleFileAttachmentSuccessAlert(validationErrors, validationWarnings);
            };

            const handleFileAttachmentErrorAlert = function (validationErrors) {
                $(`${ModalSelectors.CreateMessage.FileErrorAlert} .file-alert-content`).empty();
                if (validationErrors.length === 0) {
                    return;
                }

                $(ModalSelectors.CreateMessage.FileErrorAlert).show();

                if (validationErrors.length === 1) {
                    $(`${ModalSelectors.CreateMessage.FileErrorAlert} .file-alert-content`).text(validationErrors[0]);
                    return;
                }

                const html = `<ul style="list-style:unset;padding-left:14px;">${validationErrors.map((e) => `<li>${e}</li>`).join("")}</ul>`;
                $(`${ModalSelectors.CreateMessage.FileErrorAlert} .file-alert-content`).html(html);
            };

            const handleFileAttachmentSuccessAlert = function (validationErrors, validationWarnings) {
                $(`${ModalSelectors.CreateMessage.FileSuccessAlert} .file-alert-content`).empty();

                if (validationErrors.length === 0 && validationWarnings.length === 0) {
                    $(ModalSelectors.CreateMessage.FileSuccessAlert).show();
                    $(`${ModalSelectors.CreateMessage.FileSuccessAlert} .file-alert-content`).text("File attached successfully.");
                    return;
                }

                if (validationWarnings.length === 0) {
                    return;
                }

                $(ModalSelectors.CreateMessage.FileSuccessAlert).show();

                if (validationWarnings.length === 1) {
                    $(`${ModalSelectors.CreateMessage.FileSuccessAlert} .file-alert-content`).text(validationWarnings[0]);
                    return;
                }

                const html = `<ul style="list-style:unset;padding-left:14px;">${validationWarnings.map((e) => `<li>${e}</li>`).join("")}</ul>`;
                $(`${ModalSelectors.CreateMessage.FileSuccessAlert} .file-alert-content`).html(html);
            };

            const resetCreateMessageModal = function (subject, isReplying) {
                $(ModalSelectors.CreateMessage.Modal).modal();
                $(ModalSelectors.CreateMessage.Subject).val("").trigger("change").prop("disabled", false);
                $(ModalSelectors.CreateMessage.Content).val("").trigger("change");
                $(ModalSelectors.CreateMessage.FileInput).val("").trigger("change");
                $(ModalSelectors.CreateMessage.Form).hide();
                $(ModalSelectors.CreateMessage.FileTable).hide();
                $(`${ModalSelectors.CreateMessage.FileTable} tbody`).empty();
                $(ModalSelectors.CreateMessage.FileUploading).hide();
                $(ModalSelectors.CreateMessage.FileSuccessAlert).hide();
                $(ModalSelectors.CreateMessage.FileErrorAlert).hide();
                AttachedFiles = [];

                if (!isReplying) {
                    return;
                }

                $(ModalSelectors.CreateMessage.Subject)
                    .prop("disabled", true)
                    .val(!subject?.startsWith("RE:") ? `RE:  ${subject}` : subject || "")
                    .trigger("change");
                $(ModalSelectors.CreateMessage.Title).text("Reply to Message");
            };

            const createMessageAndUploadFiles = async function (documentPath, tag, regardingId, regardingType, userId) {
                FRCC.Portal.Utilities.showProgressIndicator("Creating message...");

                let messageId = null;
                let sanitizedContent = DOMPurify.sanitize($(ModalSelectors.CreateMessage.Content).val(), {
                    ALLOWED_TAGS: [], // Disallow all tags
                    ALLOWED_ATTR: []  // Disallow all attributes
                });
                
                try {
                    const messageData = {
                        [`eyfrcc_RegardingRecordId_${regardingType}@odata.bind`]: `/${FRCC.Portal.Utilities.getEntitySetName(regardingType)}(${regardingId})`,
                        eyfrcc_name: $(ModalSelectors.CreateMessage.Subject).val(),
                        eyfrcc_content: sanitizedContent, //$(ModalSelectors.CreateMessage.Content).val(),
                        eyfrcc_direction: Senders.Applicant,
                    };

                    messageId = await new Promise((resolve, reject) => {
                        Utilities.sendAjaxRequest(
                            {
                                type: "POST",
                                url: "/_api/eyfrcc_messages",
                                contentType: "application/json",
                                data: JSON.stringify(messageData),
                                success: function (response, status, xhr) {
                                    resolve(xhr.getResponseHeader("entityid"));
                                },
                                error: function (xhr, status, error) {
                                    reject(error);
                                },
                            },
                            "Creating message..."
                        );
                    });
                } catch (error) {
                    Utilities.handleError(error);
                }

                if (!messageId) {
                    return;
                }

                try {
                    const files = AttachedFiles.map((f) => f.File);
                    if (files && files.length > 0) {
                        FRCC.Portal.Utilities.showProgressIndicator("Uploading attachments...");

                        const fileUploadPromises = files.map((file) => {
                            const folderName = "Message Documents";
                            const now = new Date();
                            const uploadDate = `${now.getUTCMonth() + 1}/${now.getUTCDate()}/${now.getUTCFullYear()} ${now.getUTCHours() % 12 || 12}-${now
                                .getUTCMinutes()
                                .toString()
                                .padStart(2, "0")}-${now.getUTCSeconds().toString().padStart(2, "0")} ${now.getUTCHours() >= 12 ? "PM" : "AM"}`;

                            const payload = new FormData();
                            payload.append(
                                "json",
                                JSON.stringify({
                                    Name: file.name,
                                    DocumentFolder: folderName,
                                    RecordFolder: documentPath,
                                    Tags: {
                                        Id: tag,
                                        Folder: folderName,
                                        "Upload Date": uploadDate,
                                        Status: "Accepted",
                                        RegardingMessageId: messageId,
                                    },
                                    OverwriteExisting: false,
                                })
                            );
                            payload.append("file", file);

                            const eygaApiContext = new FRCC.Portal.Utilities.EygaApiContext(regardingId, regardingType, userId);
                            return FRCC.Portal.Utilities.callEygaApi(FRCC.Portal.Utilities.ApiCall.UploadDocument, eygaApiContext, payload)
                                .then((result) => ({
                                    status: "fulfilled",
                                    fileName: file.name,
                                    result,
                                }))
                                .catch((error) => ({
                                    status: "rejected",
                                    fileName: file.name,
                                    error,
                                }));
                        });

                        const uploadResults = await Promise.all(fileUploadPromises);

                        const failedUploads = uploadResults.filter((res) => res.status === "rejected");
                        if (failedUploads.length > 0) {
                            const failedFileNames = failedUploads.map((f) => `<li>${f.fileName}</li>`).join("");
                            Confirmation.openModal({
                                Title: "Error",
                                Message: `The message was sent successfully, but the following files failed to upload. Please try again:<br><ul style="list-style:unset;padding-left:14px;padding-top:14px;">${failedFileNames}</ul>`,
                                MessageContainsHTML: true,
                                ShowCancelButton: false,
                                ConfirmButtonLabel: "Close",
                                NotificationLevel: "Error",
                                ConfirmationCallback: () => Confirmation.closeModal(),
                            });
                        }
                    }
                } catch (error) {
                    Utilities.handleError(error);
                } finally {
                    FRCC.Portal.Utilities.hideProgressIndicator();
                    $(ModalSelectors.CreateMessage.Modal).modal("hide");
                    handleMessagesPageToggle(1);
                }
            };

            const openViewMessageModal = async function (id, messagesData, regardingId, regardingType, userId) {
                if (!messagesData || !messagesData[id]) {
                    return;
                }

                const message = messagesData[id];

                resetViewMessageModal();

                $(ModalSelectors.ViewMessage.Modal).modal("show");
                FRCC.Portal.Utilities.showContextualProgressIndicator("viewMessage");

                const eygaApiContext = new FRCC.Portal.Utilities.EygaApiContext(regardingId, regardingType, userId);
                const payload = {
                    RegardingMessageId: id,
                    Status: "Accepted",
                };

                const response = await FRCC.Portal.Utilities.callEygaApi(FRCC.Portal.Utilities.ApiCall.GetDocumentsById, eygaApiContext, payload);
                if (response && response.length > 0) {
                    for (const file of response) {
                        const sasUrl = await FRCC.Portal.Utilities.callEygaApi(
                            FRCC.Portal.Utilities.ApiCall.GetSASUrl,
                            eygaApiContext,
                            null,
                            `?blobName=${file.fullName}`
                        );
                        $(`${ModalSelectors.ViewMessage.DocumentContainer} ul`).append(`
                            <li>
                                <a href="${sasUrl}" target="_blank" title="Opens in a new window">${file.name}<span class="glyphicon glyphicon-new-window icon-size-xs ml-2"></span></a>
                            </li>`);
                    }
                    $(ModalSelectors.ViewMessage.DocumentContainer).show();
                }

                $(ModalSelectors.ViewMessage.DateSent).text(formatDateTime(message.SentOnDateAndTime));
                $(ModalSelectors.ViewMessage.Sender).text(message.Sender);
                $(ModalSelectors.ViewMessage.Regarding)
                    .text(`${message.ParentNumber}-${message.ParentOpportunityName}`)
                    .attr("href", `${message.DetailsPageUrl}?parentId=${message.ParentId}`);
                $(ModalSelectors.ViewMessage.Subject).text(message.Subject || "No subject");
                $(ModalSelectors.ViewMessage.Description).html(message.Description);

                FRCC.Portal.Utilities.hideContextualProgressIndicator("viewMessage");

                $(ModalSelectors.ViewMessage.ReplyButton).click(function () {
                    $(ModalSelectors.ViewMessage.Modal).modal("hide");
                    openCreateMessageModal(regardingId, regardingType, userId, message.Subject, true);
                });

                return false;
            };

            const resetViewMessageModal = function () {
                $(ModalSelectors.ViewMessage.DateSent).text("");
                $(ModalSelectors.ViewMessage.Sender).text("");
                $(ModalSelectors.ViewMessage.Regarding).text("");
                $(ModalSelectors.ViewMessage.Subject).text("");
                $(ModalSelectors.ViewMessage.Description).html("");
                $(ModalSelectors.ViewMessage.DocumentContainer).hide();
                $(`${ModalSelectors.ViewMessage.DocumentContainer} ul`).empty();
            };

            const formatDateTime = function (dateString) {
                if (!dateString) {
                    return "";
                }

                const localDateTime = new Date(dateString);
                const localDateTimeLength = localDateTime.toLocaleTimeString().length;

                const options = {
                    year: "numeric",
                    month: "2-digit",
                    day: "2-digit",
                    hour: "2-digit",
                    minute: "2-digit",
                    second: "2-digit",
                    hour12: true,
                };

                return localDateTime.toLocaleString("en-US", options).replace(",", "");
            };

            return {
                openCreateMessageModal: openCreateMessageModal,
                openViewMessageModal: openViewMessageModal,
            };
        })();

        const ApplicationForms = (function () {
            const HtmlComponentTemplates = {
                FormRow: `
                    <tr>
                        <td colspan="1" rowspan="1" class="clearfix cell lookup form-control-cell">
                            <div class="info :required:">
                                <label id=":controlId:_label" class="field-label" for=":controlId:">:controlLabel:</label>
                            </div>
                            <div class="control">
                                :controlHtml:
                            </div>
                            :descriptionHtml:
                        </td>
                        <td class="cell zero-cell" style=""></td>
                    </tr>`,
                Description: `
                    <div class="description below">
                        <span>:descriptionText:</span>
                    </div>`,
                LookupSelect: `
                    <select id=":controlId:_dropdown" class="lookup form-control" value=:value:>
                        <option label="" aria-label="Blank" selected disabled>Select a value...</option>
                        :optionsHtml:
                    </select>
                    <span class="ui"></span>`,
                Multiselect: `
                    <div class="control">
                        <div id="PcfControl_:controlId:">
                            <span id=":controlId:_Container" class="flexbox">
                                <span
                                    id="PcfControlConfig_:controlId:"
                                    data-pcf-control='{"SubscribedFunctionalities":{},"DataConnectors":[],"AuthConfigPropertiesGroups":[],"Resources":[{"Type":3,"Path":"{\\"Id\\":\\"ManifestTemplate\\",\\"Data\\":{\\"ConstructorName\\":\\"MscrmControls.MultiSelectPicklist.UpdMSPicklistControl\\",\\"CustomControlId\\":\\"MscrmControls.MultiSelectPicklist.UpdMSPicklistControl\\",\\"IsVirtual\\":false,\\"Properties\\":{\\"Resources\\":[],\\"GroupDefinitions\\":null,\\"DataSetDefinitions\\":{},\\"Properties\\":{\\"value\\":{\\"Required\\":true,\\"Usage\\":0,\\"DefaultValue\\":\\"\\"}},\\"FeatureUsage\\":null},\\"Events\\":null,\\"CommonEvents\\":null,\\"DesignMap\\":null,\\"PropertyDependencies\\":[]}}","ModifiedPath":"","LoadingOrder":1,"IsControlSpecific":false},{"Type":3,"Path":"{\\"Id\\":\\"ManifestPropertyTypes\\",\\"Data\\":{\\"Properties\\":[{\\"Name\\":\\"value\\",\\"Type\\":\\"MultiSelectOptionSet\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":false},{\\"Name\\":\\"X\\",\\"Type\\":\\"Number\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":true},{\\"Name\\":\\"Y\\",\\"Type\\":\\"Number\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":true},{\\"Name\\":\\"Width\\",\\"Type\\":\\"Number\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":true},{\\"Name\\":\\"Height\\",\\"Type\\":\\"Number\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":true},{\\"Name\\":\\"Visible\\",\\"Type\\":\\"Boolean\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":true},{\\"Name\\":\\"DisplayMode\\",\\"Type\\":\\"DisplayMode\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":true},{\\"Name\\":\\"TabIndex\\",\\"Type\\":\\"Number\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":true},{\\"Name\\":\\"Tooltip\\",\\"Type\\":\\"String\\",\\"ComplexType\\":\\"\\",\\"IsPrimary\\":false,\\"IsIncluded\\":true}]}}","ModifiedPath":"","LoadingOrder":2,"IsControlSpecific":false},{"Type":3,"Path":"{\\"Id\\":\\"PowerAppsControlInfo\\",\\"Data\\":{\\"TemplateVersion\\":1,\\"Version\\":\\"1.0.67\\",\\"IsFirstParty\\":false}}","ModifiedPath":"","LoadingOrder":3,"IsControlSpecific":false},{"Type":3,"Path":"{\\"Id\\":\\"FormattingData\\",\\"Data\\":{\\"timeZoneUtcOffsetMinutes\\":0,\\"dateTimeFormat\\":{\\"AMDesignator\\":\\"AM\\",\\"Calendar\\":{\\"MinSupportedDateTime\\":\\"0001-01-01T00:00:00\\",\\"MaxSupportedDateTime\\":\\"9999-12-31T23:59:59.9999999\\",\\"AlgorithmType\\":1,\\"CalendarType\\":1,\\"Eras\\":[1],\\"TwoDigitYearMax\\":2049,\\"IsReadOnly\\":false},\\"DateSeparator\\":\\"/\\",\\"FirstDayOfWeek\\":0,\\"CalendarWeekRule\\":0,\\"FullDateTimePattern\\":\\"dddd, MMMM d, yyyy h:mm:ss tt\\",\\"LongDatePattern\\":\\"dddd, MMMM d, yyyy\\",\\"LongTimePattern\\":\\"h:mm:ss tt\\",\\"MonthDayPattern\\":\\"MMMM d\\",\\"PMDesignator\\":\\"PM\\",\\"RFC1123Pattern\\":\\"ddd, dd MMM yyyy HH&apos;:&apos;mm&apos;:&apos;ss &apos;GMT&apos;\\",\\"ShortDatePattern\\":\\"M/d/yyyy\\",\\"ShortTimePattern\\":\\"h:mm tt\\",\\"SortableDateTimePattern\\":\\"yyyy&apos;-&apos;MM&apos;-&apos;dd&apos;T&apos;HH&apos;:&apos;mm&apos;:&apos;ss\\",\\"TimeSeparator\\":\\":\\",\\"UniversalSortableDateTimePattern\\":\\"yyyy&apos;-&apos;MM&apos;-&apos;dd HH&apos;:&apos;mm&apos;:&apos;ss&apos;Z&apos;\\",\\"YearMonthPattern\\":\\"MMMM yyyy\\",\\"AbbreviatedDayNames\\":[\\"Sun\\",\\"Mon\\",\\"Tue\\",\\"Wed\\",\\"Thu\\",\\"Fri\\",\\"Sat\\"],\\"ShortestDayNames\\":[\\"Su\\",\\"Mo\\",\\"Tu\\",\\"We\\",\\"Th\\",\\"Fr\\",\\"Sa\\"],\\"DayNames\\":[\\"Sunday\\",\\"Monday\\",\\"Tuesday\\",\\"Wednesday\\",\\"Thursday\\",\\"Friday\\",\\"Saturday\\"],\\"AbbreviatedMonthNames\\":[\\"Jan\\",\\"Feb\\",\\"Mar\\",\\"Apr\\",\\"May\\",\\"Jun\\",\\"Jul\\",\\"Aug\\",\\"Sep\\",\\"Oct\\",\\"Nov\\",\\"Dec\\",\\"\\"],\\"MonthNames\\":[\\"January\\",\\"February\\",\\"March\\",\\"April\\",\\"May\\",\\"June\\",\\"July\\",\\"August\\",\\"September\\",\\"October\\",\\"November\\",\\"December\\",\\"\\"],\\"IsReadOnly\\":false,\\"NativeCalendarName\\":\\"Gregorian Calendar\\",\\"AbbreviatedMonthGenitiveNames\\":[\\"Jan\\",\\"Feb\\",\\"Mar\\",\\"Apr\\",\\"May\\",\\"Jun\\",\\"Jul\\",\\"Aug\\",\\"Sep\\",\\"Oct\\",\\"Nov\\",\\"Dec\\",\\"\\"],\\"MonthGenitiveNames\\":[\\"January\\",\\"February\\",\\"March\\",\\"April\\",\\"May\\",\\"June\\",\\"July\\",\\"August\\",\\"September\\",\\"October\\",\\"November\\",\\"December\\",\\"\\"]},\\"numberFormat\\":{\\"CurrencyDecimalDigits\\":2,\\"CurrencyDecimalSeparator\\":\\".\\",\\"IsReadOnly\\":false,\\"CurrencyGroupSizes\\":[3],\\"NumberGroupSizes\\":[3],\\"PercentGroupSizes\\":[3],\\"CurrencyGroupSeparator\\":\\",\\",\\"CurrencySymbol\\":\\"$\\",\\"NaNSymbol\\":\\"NaN\\",\\"CurrencyNegativePattern\\":0,\\"NumberNegativePattern\\":1,\\"PercentPositivePattern\\":1,\\"PercentNegativePattern\\":1,\\"NegativeInfinitySymbol\\":\\"-∞\\",\\"NegativeSign\\":\\"-\\",\\"NumberDecimalDigits\\":2,\\"NumberDecimalSeparator\\":\\".\\",\\"NumberGroupSeparator\\":\\",\\",\\"CurrencyPositivePattern\\":0,\\"PositiveInfinitySymbol\\":\\"∞\\",\\"PositiveSign\\":\\"+\\",\\"PercentDecimalDigits\\":2,\\"PercentDecimalSeparator\\":\\".\\",\\"PercentGroupSeparator\\":\\",\\",\\"PercentSymbol\\":\\"%\\",\\"PerMilleSymbol\\":\\"‰\\",\\"NativeDigits\\":[\\"0\\",\\"1\\",\\"2\\",\\"3\\",\\"4\\",\\"5\\",\\"6\\",\\"7\\",\\"8\\",\\"9\\"],\\"DigitSubstitution\\":1},\\"timeZoneAdjusters\\":[],\\"formatInfoCultureName\\":\\"1033\\",\\"formatter\\":null,\\"languagesByCode\\":{\\"1033\\":\\"English (United States)\\",\\"-1\\":\\"All Languages\\"}}}","ModifiedPath":"","LoadingOrder":54,"IsControlSpecific":false},{"Type":3,"Path":"{\\"Id\\":\\"LocalizedStrings\\",\\"Data\\":{\\"LocStrings\\":{\\"MultiSelectPicklistControl.DeleteItemLabel\\":\\"Remove {0}\\",\\"MultiSelectPicklistControl.SelectedItems\\":\\"Selected:\\",\\"MultiSelectPicklistControl.Collapse\\":\\"less\\",\\"MultiSelectPicklistControl.ResultsFound\\":\\"Number of results: {0}\\",\\"MultiSelectPicklistControl.DropdownToggleLabel\\":\\"Toggle menu\\",\\"MultiSelectPicklistControl.NoItemsAvailable\\":\\"No entries found\\",\\"MultiSelectPicklistControl.ExpandLabel\\":\\"Show {0} more selected items\\",\\"MultiSelectPicklistControl.ComboboxInstructions\\":\\"Use the arrow keys to navigate through the list of items. Press space to select or deselect an item. For touch devices, use swipe and touch gestures to explore this list.\\",\\"MultiSelectPicklistControl.ViewmodeExpandLabel\\":\\"Expand\\",\\"MultiSelectPicklistControl.SearchPlaceholder\\":\\"Select or search options\\",\\"MultiSelectPicklistControl.TotalItems\\":\\"{0} items\\",\\"MultiSelectPicklistControl.SelectAllDeselect\\":\\"{0} items removed\\",\\"MultiSelectPicklistControl.ViewmodeExpandCollapseLabel\\":\\"Expand or collapse selected content\\",\\"MultiSelectPicklistControl.ViewmodeCollapseLabel\\":\\"Collapse\\",\\"MultiSelectPicklistControl.SelectedItemsLabel\\":\\"Selected values for {0}\\",\\"MultiSelectPicklistControl.SelectedItemsInstructions\\":\\"Use the arrow keys to navigate through the list of currently selected items. Press delete or backspace to deselect an item. For touch devices, use swipe and touch gestures to explore the list.\\",\\"MultiSelectPicklistControl.SelectedItemAdded\\":\\"{0} added\\",\\"MultiSelectPicklistControl.SelectAllSelect\\":\\"{0} items selected\\",\\"MultiSelectPicklistControl.SelectAll\\":\\"Select all\\",\\"MultiSelectPicklistControl.Expand\\":\\"+{0}\\",\\"MultiSelectPicklistControl.SelectedItemRemoved\\":\\"{0} removed\\",\\"MultiSelectPicklistControl.CollapseLabel\\":\\"Show fewer selected items\\"}}}","ModifiedPath":"","LoadingOrder":55,"IsControlSpecific":false}],"DataSets":[],"Properties":[{"Name":"value","DisplayName":"CC_Field","Description":"CC_Field_Desc","Type":52,"Usage":2,"Hidden":true,"DefaultValue":"","PhoneDefaultValue":"","WebDefaultValue":"","NullDefaultValue":null,"HelperUI":"","Category":"data","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":[],"Required":true,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"CC_Field","DescriptionResourceString":"CC_Field_Desc"}],"IncludedProperties":[{"Name":"X","DisplayName":"CC_X","Description":"","Type":2,"Usage":2,"Hidden":false,"DefaultValue":"","PhoneDefaultValue":"","WebDefaultValue":"","NullDefaultValue":"","HelperUI":"","Category":"design","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":null,"Required":false,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"CC_X","DescriptionResourceString":""},{"Name":"Y","DisplayName":"CC_Y","Description":"","Type":2,"Usage":2,"Hidden":false,"DefaultValue":"","PhoneDefaultValue":"","WebDefaultValue":"","NullDefaultValue":"","HelperUI":"","Category":"design","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":null,"Required":false,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"CC_Y","DescriptionResourceString":""},{"Name":"Width","DisplayName":"Width","Description":"","Type":2,"Usage":2,"Hidden":false,"DefaultValue":"200","PhoneDefaultValue":"350","WebDefaultValue":"112","NullDefaultValue":"0","HelperUI":"","Category":"design","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":null,"Required":false,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"Width","DescriptionResourceString":""},{"Name":"Height","DisplayName":"Height","Description":"","Type":2,"Usage":2,"Hidden":false,"DefaultValue":"200","PhoneDefaultValue":"350","WebDefaultValue":"112","NullDefaultValue":"0","HelperUI":"","Category":"design","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":null,"Required":false,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"Height","DescriptionResourceString":""},{"Name":"Visible","DisplayName":"Visible","Description":"","Type":4,"Usage":2,"Hidden":false,"DefaultValue":"true","PhoneDefaultValue":"","WebDefaultValue":"","NullDefaultValue":"","HelperUI":"","Category":"design","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":null,"Required":false,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"Visible","DescriptionResourceString":""},{"Name":"DisplayMode","DisplayName":"DisplayMode","Description":"","Type":29,"Usage":2,"Hidden":false,"DefaultValue":"","PhoneDefaultValue":"","WebDefaultValue":"","NullDefaultValue":"","HelperUI":"","Category":"design","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":null,"Required":false,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"DisplayMode","DescriptionResourceString":""},{"Name":"TabIndex","DisplayName":"TabIndex","Description":"","Type":2,"Usage":0,"Hidden":false,"DefaultValue":"0","PhoneDefaultValue":"","WebDefaultValue":"","NullDefaultValue":"","HelperUI":"","Category":"design","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":null,"Required":false,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"TabIndex","DescriptionResourceString":""},{"Name":"Tooltip","DisplayName":"Tooltip","Description":"","Type":3,"Usage":0,"Hidden":false,"DefaultValue":"CC_MultiSelectPicklist_Name","PhoneDefaultValue":"","WebDefaultValue":"","NullDefaultValue":"","HelperUI":"","Category":"design","IsPrimaryBehavioral":false,"IsPrimaryInput":false,"IsPrimaryOutput":false,"IsResettable":false,"EnumValues":null,"Required":false,"IsDataSourceProperty":false,"PassThroughProperty":"","ShouldAutoBind":false,"ParentDataSetName":"","DisplayNameResourceString":"Tooltip","DescriptionResourceString":""}],"PropertyDependencies":[],"Events":[],"CommonEvents":[{"Name":"OnChange","DisplayName":"OnChange","Description":"","Hidden":false}],"Conversions":[],"Name":"MscrmControls.MultiSelectPicklist.UpdMSPicklistControl","IsCdnControl":true}'
                                    pcf-controlcontext=':pcfControlContext:'
                                ></span>
                                <div id=":controlId:_ControlView" class="flexbox"></div>
                            </span>
                        </div>
                        <input type="hidden" id=":controlId:" value='[]' />
                    </div>`,
                PCFControlContext: `{"UniqueId":":controlId:_ControlView","ControlId":":controlId:","GridDataControlId":null,"ControlName":"MscrmControls.MultiSelectPicklist.UpdMSPicklistControl","ControlProperties":{"value_Options"::valueAllOptions:,"value_DefaultValue":null,"value_DisplayName":":controlLabel:","value":[]},"IsDisabled":false,"PcfControlHiddenValueId":":controlId:","PcfDynamicProperyName":"value"}`,
                EmptyMultiselectOption: `
                    <li class="msos-option msos-hidden" role="option" style="" aria-selected="false"></li>`,
                OutcomeInput: `
                    <input type="text" data-outcomeid=":controlId:" id="eyfrcc_outcomevalue-:controlId:" class="text integer form-control output-value-input" required>`,
                SectionHeader: ``,
            };

            let Configuration;

            const initializeApplicationFormModals = function (configuration) {
                Configuration = configuration;

                initializeSubapplicantRecompetingNewModal();
            };

            const initializeSubapplicantRecompetingNewModal = function () {
                const $iFrame = $(`#eyfrcc_subrecipient_subgrid iframe[data-page]`);
                $iFrame.on("load", loadSubapplicantRecompetingNewModal);
            };

            const loadSubapplicantRecompetingNewModal = async function (e) {
                const formSource = $(this).attr("src");
                if (!formSource) {
                    return;
                }

                if (!Configuration.ApplicationTypeId) {
                    return;
                }

                const iFrameContentWindow = e.currentTarget.contentWindow;
                const iFrame$ = iFrameContentWindow.$;
                const $fundingOpportunity = iFrame$("#eyfrcc_fundingopportunity");
                if ($fundingOpportunity.length === 0) {
                    return;
                }

                FRCC.Portal.Utilities.showProgressIndicator("Loading Modal...");

                const fundingOpportunitiesResponse = await FRCC.Portal.Utilities.asyncAppAjax({
                    type: "GET",
                    url: `/Services/GetFundingOpportunitiesForRecompetingNewSubapplicant?applicationTypeId=${Configuration.ApplicationTypeId}`,
                    contentType: "application/json",
                });
                const fundingOpportunities = JSON.parse(fundingOpportunitiesResponse)?.results;

                const optionsListHtml = Utilities.formatOptionsListHtml(fundingOpportunities);
                Utilities.createLookupDropdown(iFrameContentWindow, iFrame$, $fundingOpportunity, optionsListHtml, "eyfrcc_applicationtype", true);

                FRCC.Portal.Utilities.hideProgressIndicator();
            };

            return {
                initializeApplicationFormModals: initializeApplicationFormModals,
                HtmlComponentTemplates: HtmlComponentTemplates,
            };
        })();

        const Utilities = (function () {
            const sendAjaxRequest = function (options, progressMessage) {
                FRCC.Portal.Utilities.showProgressIndicator(progressMessage);

                _safeAjax({
                    type: options.type,
                    url: options.url,
                    contentType: "application/json",
                    data: options.data,
                    success: function (response, status, xhr) {
                        if (options.success) {
                            options.success(response, status, xhr);
                            return;
                        }
                        hideModalsAndReload();
                    },
                    error: function (request, status, error) {
                        handleError(request.responseJSON.error.innererror?.message || request.responseJSON.error.message);
                        if (options.error) {
                            options.error(request, status, error);
                        }
                    },
                });
            };

            const _safeAjax = function (ajaxOptions) {
                const deferredAjax = $.Deferred();
                shell
                    .getTokenDeferred()
                    .done(function (token) {
                        if (!ajaxOptions.headers) {
                            $.extend(ajaxOptions, {
                                headers: {
                                    __RequestVerificationToken: token,
                                },
                            });
                        } else {
                            ajaxOptions.headers["__RequestVerificationToken"] = token;
                        }
                        $.ajax(ajaxOptions)
                            .done(function (data, textStatus, jqXHR) {
                                validateLoginSession(data, textStatus, jqXHR, deferredAjax.resolve);
                            })
                            .fail(deferredAjax.reject);
                    })
                    .fail(function () {
                        deferredAjax.rejectWith(this, arguments);
                    });
                return deferredAjax.promise();
            };

            const handleError = function (message, html = null) {
                FRCC.Portal.Utilities.hideProgressIndicator();

                const overwriteMessage = FRCC.Portal.Utilities.overwriteErrorMessage(message);
                if (overwriteMessage !== null && !overwriteMessage.isHtml) {
                    message = overwriteMessage.updatedMessage;
                }

                $(".modal-body-error-container").show();
                $(".modal-body-error-container .modal-body-description span").remove(); // Remove any previous html contained within modal-body-description
                $(".modal .modal-body").scrollTop(0);

                if (html != null || (overwriteMessage !== null && overwriteMessage.isHtml)) {
                    $(".modal-body-error-container .modal-body-description").text(null);
                    $(".modal-body-error-container .modal-body-description").append(html ?? overwriteMessage.updatedMessage);
                    return;
                }

                $(".modal-body-error-container .modal-body-description").text(message);
            };

            const clearValidationErrors = function () {
                $(".has-error").removeClass("has-error");
                $(".modal-body-error-container").hide();
            };

            const hideModalsAndReload = function () {
                FRCC.Portal.Utilities.hideProgressIndicator();
                $(".modal").modal("hide");
                location.reload();
            };

            const formatDateInput = function (inputDateString) {
                if (!inputDateString) {
                    return null;
                }

                const inputDate = new Date(inputDateString);

                const year = inputDate.getFullYear();
                const month = ("0" + (inputDate.getMonth() + 1)).slice(-2);
                const day = ("0" + inputDate.getDate()).slice(-2);

                return `${year}-${month}-${day}`;
            };

            const formatOptionsListHtml = function (records) {
                return records.map((record) => `<option value="${record.Id}">${record.Name}</option>`).join("");
            };

            const createLookupDropdown = function (iFrameContentWindow, iFrame$, $lookupControl, optionList, entityName, defaultValue = false) {
                const selectedValue = $lookupControl.val();
                const lookupControlId = $lookupControl.attr("id");
                const lookupControlRow = $lookupControl.closest("tr");
                const clonedRow = lookupControlRow.clone();

                lookupControlRow.children("td.cell").removeClass("cell").addClass("hidden-cell").hide();
                const controlCell = clonedRow.find(".control").empty();

                const templateParameters = {
                    controlId: lookupControlId,
                    value: selectedValue,
                    optionsHtml: optionList,
                };
                const selectHtml = FRCC.AmeriCorp.Utilities.generateTemplate(ApplicationForms.HtmlComponentTemplates.LookupSelect, templateParameters);

                controlCell.append(selectHtml);
                lookupControlRow.after(clonedRow);

                iFrame$(`#${lookupControlId}_dropdown`).val(selectedValue);

                iFrame$(`div.info label[id='${lookupControlId}_label']`).attr("for", `${lookupControlId}_dropdown`);
                iFrame$(`#${lookupControlId}_dropdown`).change(function () {
                    const $dropdown = iFrame$(this);
                    const value = $dropdown.val();

                    iFrame$(`input#${lookupControlId}`).val(value ?? "");
                    iFrame$(`input#${lookupControlId}_entityname`).val(value ? entityName : "");
                });

                replaceSubmitValidation(iFrameContentWindow, lookupControlId, `${lookupControlId}_dropdown`);

                if (defaultValue) {
                    defaultLookupDropdown(iFrame$, lookupControlId, selectedValue);
                }
            };

            const replaceSubmitValidation = function (iFrameContentWindow, targetField, replaceFieldLogicalName) {
                iFrameContentWindow.Page_Validators.forEach((validator, index, array) => {
                    if (!validator.id.contains(targetField)) {
                        return;
                    }

                    array[index].errormessage = validator.errormessage.replace(`"${targetField}"`, `"${replaceFieldLogicalName}"`);
                });
            };

            const defaultLookupDropdown = function (iFrame$, lookupControlId, selectedValue) {
                if (selectedValue) {
                    return;
                }

                const options = iFrame$(`#${lookupControlId}_dropdown option[value]`);
                if (options.length !== 1) {
                    return;
                }

                iFrame$(`#${lookupControlId}_dropdown`).val(iFrame$(`#${lookupControlId}_dropdown option[value]:first`).val());
                iFrame$(`#${lookupControlId}_dropdown`).change();
            };

            const generateGuid = function () {
                const randomGuid = "xxxxxxxx-xxxx-xxxx-xxxx-xxxxxxxxxxxx".replace(/x/g, function (c) {
                    const r = (Math.random() * 16) | 0,
                        v = c === "x" ? r : (r & 0x3) | 0x8;
                    return v.toString(16);
                });

                return randomGuid;
            };

            return {
                sendAjaxRequest: sendAjaxRequest,
                handleError: handleError,
                clearValidationErrors: clearValidationErrors,
                formatDateInput: formatDateInput,
                formatOptionsListHtml: formatOptionsListHtml,
                createLookupDropdown: createLookupDropdown,
                generateGuid: generateGuid,
            };
        })();

        const Validations = (function () {
            const validateForm = function (modalSelector, validateRequiredFields = true, customValidations = []) {
                clearValidationErrors();
                let isValid = true;
                let validationsHtml = [];

                if (validateRequiredFields) {
                    const $requiredFields = $(`${modalSelector} label.required`).next("input, textarea, select").filter(":visible");
                    $requiredFields.each(function () {
                        const $field = $(this);
                        if (!$field.val() || !$field.val().trim()) {
                            isValid = false;
                            const fieldId = $field.attr("id");
                            const fieldLabel = $(`label[for="${fieldId}"]`).text();
                            validationsHtml.push(`<li><a href="#${fieldId}">${fieldLabel} is a required field.</a></li>`);
                            $field.closest(".form-group").addClass("has-error");
                        }
                    });
                }

                customValidations.forEach((customValidation) => {
                    if (!customValidation.EvaluationFunction()) {
                        isValid = false;
                        validationsHtml.push(`<li><a href="${customValidation.TargetField}">${customValidation.Message}</a></li>`);
                        $(customValidation.TargetField).closest(".form-group").addClass("has-error");
                    }
                });

                if (!isValid) {
                    const html = `<ul>${validationsHtml.join("")}</ul>`;
                    Utilities.handleError(null, html);
                    $(`${modalSelector} .modal-body`).scrollTop(0);
                }

                return isValid;
            };

            const clearValidationErrors = function () {
                $(".has-error").removeClass("has-error");
                $(".modal-body-error-container").hide();
            };

            const validateZipCode = function (zipCode) {
                const pattern = /^\d{5}-\d{4}$/;
                return pattern.test(zipCode);
            };

            const validateEmailAddress = function (email) {
                const pattern = /^[^\s@]+@[^\s@]+\.[^\s@]+$/;
                return pattern.test(email);
            };

            return {
                validateForm: validateForm,
                validateZipCode: validateZipCode,
                validateEmailAddress: validateEmailAddress,
            };
        })();

        return {
            NegotiatedIndirectCostRateAgreement: NegotiatedIndirectCostRateAgreement,
            RateLine: RateLine,
            Project: Project,
            SessionManagement: SessionManagement,
            LateSubmission: LateSubmission,
            Confirmation: Confirmation,
            AddressValidation: AddressValidation,
            Messages: Messages,
            ApplicationForms: ApplicationForms,
            Utilities: Utilities,
            Validations: Validations,
        };
    })();
})(this);
